/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "client.h"
#include "qmenu.h"

#define REF_SOFT	0
#define REF_OPENGL	1
#define REF_3DFX	2
#define REF_POWERVR	3
#define REF_VERITE	4

extern cvar_t *vid_fullscreen;
extern cvar_t *vid_vsync;
extern cvar_t *vid_gamma;
extern cvar_t *r_uiscale;

static cvar_t *vid_mode;
static cvar_t *gl_picmip;
static cvar_t *gl_finish;

static cvar_t *sw_mode;
static cvar_t *sw_stipplealpha;

extern cvar_t *gl_textureanisotropy;

extern void M_ForceMenuOff (void);
int VID_BoundMode (int modenum);

int Menu_DecodeOverbright (cvar_t *r_obcvar)
{
	if (r_obcvar->value < 1.5f)
		return 0;
	else if (r_obcvar->value < 3)
		return 1;
	else return 2;
}


int Menu_DecodeAF (char **amodelist)
{
	int af = 1;
	int i = 0;

	for (i = 0, af = 1;; i++, af <<= 1)
	{
		// end of the list and not found so break out
		if (!amodelist[i])
		{
			i = 0;
			break;
		}

		if (gl_textureanisotropy->value < (af * 2)) break;
	}

	return i;
}


/*
====================================================================

MENU INTERACTION

====================================================================
*/
static menuframework_s	s_opengl_menu;
static int				s_opengl_menu_index;

static menulist_s		s_mode_list;
static menulist_s		s_ref_list;
static menuslider_s		s_tq_slider;
static menuslider_s		s_uiscale_slider;
static menuslider_s		s_brightness_slider;
static menulist_s		s_texture_overbright_list;
static menulist_s		s_lightmap_overbright_list;
static menulist_s		s_anisotropic_filter_list;
static menulist_s		s_screenshot_format_list;
static menuslider_s		s_consize_slider;
static menulist_s  		s_fs_box;
static menulist_s  		s_vs_box;
static menulist_s  		s_stipple_box;
static menulist_s  		s_paletted_texture_box;
static menulist_s  		s_finish_box;
static menuaction_s		s_cancel_action;
static menuaction_s		s_apply_action;

static void DriverCallback (void *unused)
{
	s_opengl_menu_index = 1;
}

static void ScreenSizeCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;

	Cvar_SetValue ("r_uiscale", slider->curvalue * 0.25f);

	// reposition the menu because this can happen while in it
	if (r_uiscale->value > 1)
	{
		d3d_2DMode.Width = (float) viddef.Width / r_uiscale->value;
		d3d_2DMode.Height = (float) viddef.Height / r_uiscale->value;
	}
	else
	{
		d3d_2DMode.Width = viddef.Width;
		d3d_2DMode.Height = viddef.Height;
	}

	s_opengl_menu.x = d3d_2DMode.Width * 0.50;
	Menu_Center (&s_opengl_menu);
	s_opengl_menu.x -= 8;
}


extern D3DDISPLAYMODE viddef;
extern D3DDISPLAYMODE d3d_2DMode;

static void ConsizeCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;

	// fix up all the fiddle crap we need
	Menu_Center (&s_opengl_menu);
	VID_MenuInit ();
}


static void BrightnessCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;

	float gamma = (0.8 - (slider->curvalue / 10.0 - 0.5)) + 0.5;

	Cvar_SetValue ("vid_gamma", gamma);
}

static void TexOverbrightCallback (void *s)
{
	menulist_s *ml = (menulist_s *) s;

	if (ml->curvalue == 0)
		Cvar_SetValue ("intensity", 0);
	else if (ml->curvalue == 1)
		Cvar_SetValue ("intensity", 2);
	else Cvar_SetValue ("intensity", 4);
}


static void LMOverbrightCallback (void *s)
{
}


static void AFCallback (void *s)
{
	menulist_s *ml = (menulist_s *) s;

	if (ml->curvalue == 0)
		Cvar_SetValue ("gl_textureanisotropy", 1);
	else
	{
		int af = 2;
		int i = 1;

		for (i = 1, af = 2;; i++, af <<= 2)
		{
			if (ml->curvalue == i)
			{
				Cvar_SetValue ("gl_textureanisotropy", af);
				break;
			}
		}
	}
}


static void ApplyChanges (void *unused)
{
	float gamma;

	// invert sense so greater = brighter, and scale to a range of 0.5 to 1.3
	gamma = (0.8 - (s_brightness_slider.curvalue / 10.0 - 0.5)) + 0.5;

	Cvar_SetValue ("vid_gamma", gamma);
	Cvar_SetValue ("vid_fullscreen", s_fs_box.curvalue);
	Cvar_SetValue ("vid_vsync", s_vs_box.curvalue);
	Cvar_SetValue ("vid_mode", s_mode_list.curvalue);

	M_ForceMenuOff ();
}

static void CancelChanges (void *unused)
{
	extern void M_PopMenu (void);

	M_PopMenu ();
}

/*
** VID_MenuInit
*/
void D3D_PopulateVideoMenu (char ***vmodelist, char ***rrlist);
void D3D_BuildAnisotropicFilterModes (char ***amodelist);

char **resolutions = NULL;
char **refreshrates = NULL;
char **anisomodes = NULL;

static void DrawVidModeFunc (void *self)
{
	menulist_s *s = (menulist_s *) self;

	Menu_DrawString (
		s->x + s->parent->x + RCOLUMN_OFFSET,
		s->y + s->parent->y + 10,
		refreshrates[s_mode_list.curvalue]
	);
}


void VID_MenuInit (void)
{
	static char *yesno_names[] =
	{
		"no",
		"yes",
		0
	};

	// popoulate resolutions list first time in here
	if (!resolutions) D3D_PopulateVideoMenu (&resolutions, &refreshrates);
	if (!anisomodes) D3D_BuildAnisotropicFilterModes (&anisomodes);

	if (!gl_picmip) gl_picmip = Cvar_Get ("gl_picmip", "0", 0);
	if (!vid_mode) vid_mode = Cvar_Get ("vid_mode", "0", 0);
	if (!sw_mode) sw_mode = Cvar_Get ("sw_mode", "0", 0);
	if (!gl_finish) gl_finish = Cvar_Get ("gl_finish", "0", CVAR_ARCHIVE);
	if (!sw_stipplealpha) sw_stipplealpha = Cvar_Get ("sw_stipplealpha", "0", CVAR_ARCHIVE);
	if (!r_uiscale) r_uiscale = Cvar_Get ("r_uiscale", "1", CVAR_ARCHIVE);

	s_mode_list.curvalue = VID_BoundMode ((int) vid_mode->value);

	s_opengl_menu.x = d3d_2DMode.Width * 0.50;
	s_opengl_menu.nitems = 0;

	s_mode_list.type = MTYPE_SPINCONTROL;
	s_mode_list.name = "video mode";
	s_mode_list.x = 0;
	s_mode_list.y = 0;
	s_mode_list.itemnames = resolutions;
	s_mode_list.ownerdraw = DrawVidModeFunc;

	s_fs_box.type = MTYPE_SPINCONTROL;
	s_fs_box.x	= 0;
	s_fs_box.y	= 30;
	s_fs_box.name	= "fullscreen";
	s_fs_box.itemnames = yesno_names;
	s_fs_box.curvalue = !!vid_fullscreen->value;

	s_vs_box.type = MTYPE_SPINCONTROL;
	s_vs_box.x	= 0;
	s_vs_box.y	= 40;
	s_vs_box.name	= "vertical sync";
	s_vs_box.itemnames = yesno_names;
	s_vs_box.curvalue = !!vid_vsync->value;

	s_brightness_slider.type	= MTYPE_SLIDER;
	s_brightness_slider.x	= 0;
	s_brightness_slider.y	= 60;
	s_brightness_slider.name	= "brightness";
	s_brightness_slider.callback = BrightnessCallback;
	s_brightness_slider.minvalue = 5;
	s_brightness_slider.maxvalue = 13;
	s_brightness_slider.curvalue = (1.3 - vid_gamma->value + 0.5) * 10;

	s_uiscale_slider.type	= MTYPE_SLIDER;
	s_uiscale_slider.x		= 0;
	s_uiscale_slider.y		= 70;
	s_uiscale_slider.name	= "ui scale";
	s_uiscale_slider.minvalue = 4;
	s_uiscale_slider.maxvalue = 16;
	s_uiscale_slider.curvalue = r_uiscale->value * 4;
	s_uiscale_slider.callback = ScreenSizeCallback;

	s_anisotropic_filter_list.type = MTYPE_SPINCONTROL;
	s_anisotropic_filter_list.x	= 0;
	s_anisotropic_filter_list.y	= 80;
	s_anisotropic_filter_list.name	= "anisotropic filter";
	s_anisotropic_filter_list.itemnames = anisomodes;
	s_anisotropic_filter_list.curvalue = Menu_DecodeAF (anisomodes);
	s_anisotropic_filter_list.callback = AFCallback;

	s_apply_action.type = MTYPE_ACTION;
	s_apply_action.name = "apply";
	s_apply_action.x    = 0;
	s_apply_action.y    = 100;
	s_apply_action.callback = ApplyChanges;

	s_cancel_action.type = MTYPE_ACTION;
	s_cancel_action.name = "cancel";
	s_cancel_action.x    = 0;
	s_cancel_action.y    = 110;
	s_cancel_action.callback = CancelChanges;

	Menu_AddItem (&s_opengl_menu, (void *) &s_mode_list);
	Menu_AddItem (&s_opengl_menu, (void *) &s_fs_box);
	Menu_AddItem (&s_opengl_menu, (void *) &s_vs_box);
	Menu_AddItem (&s_opengl_menu, (void *) &s_brightness_slider);
	Menu_AddItem (&s_opengl_menu, (void *) &s_uiscale_slider);
	Menu_AddItem (&s_opengl_menu, (void *) &s_anisotropic_filter_list);

	Menu_AddItem (&s_opengl_menu, (void *) &s_apply_action);
	Menu_AddItem (&s_opengl_menu, (void *) &s_cancel_action);

	Menu_Center (&s_opengl_menu);
	s_opengl_menu.x -= 8;
}

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	int w, h;

	/*
	** draw the banner
	*/
	Draw_GetPicSize (&w, &h, "m_banner_video");
	Draw_Pic (d3d_2DMode.Width / 2 - w / 2, d3d_2DMode.Height / 2 - 110, "m_banner_video");

	/*
	** move cursor to a reasonable starting position
	*/
	Menu_AdjustCursor (&s_opengl_menu, 1);

	/*
	** draw the menu
	*/
	Menu_Draw (&s_opengl_menu);
}

/*
================
VID_MenuKey
================
*/
void M_PopMenu (void);

const char *VID_MenuKey (int key)
{
	menuframework_s *m = &s_opengl_menu;
	static const char *sound = "misc/menu1.wav";
	static const char *menu_out_sound = "misc/menu3.wav";

	switch (key)
	{
	case K_ESCAPE:
		M_PopMenu ();
		return menu_out_sound;
	case K_KP_UPARROW:
	case K_UPARROW:
		m->cursor--;
		Menu_AdjustCursor (m, -1);
		break;
	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		m->cursor++;
		Menu_AdjustCursor (m, 1);
		break;
	case K_KP_LEFTARROW:
	case K_LEFTARROW:
		Menu_SlideItem (m, -1);
		break;
	case K_KP_RIGHTARROW:
	case K_RIGHTARROW:
		Menu_SlideItem (m, 1);
		break;
	case K_KP_ENTER:
	case K_ENTER:

		if (!Menu_SelectItem (m))
			ApplyChanges (NULL);

		break;
	}

	return sound;
}


